import math
import numpy as np

from multiprocessing import Pool, cpu_count
from numpy.random import default_rng
from scipy.integrate import odeint


def num_samples(epsilon, delta, n_x=3, const=None):
    """Compute the number of samples needed to satisfy the specified probabilistic guarantees

    :param epsilon: The accuracy parameter
    :type epsilon: float
    :param delta: The confidence parameter
    :type delta: float
    :param n_x: The state dimension, defaults to 3
    :type n_x: int
    :param const: The constraints placed on the parameters A and b, defaults to None
    :type const: string, optional
    :return: The number of samples needed to satisfy the specified probabilistic guarantees
    :rtype: int
    """
    if const is None:
        n_theta = 0.5 * (n_x ** 2 + 3 * n_x)
    elif const == "diagonal":
        n_theta = 2 * n_x
    elif const == "scalar":
        n_theta = 1
    N = math.ceil(math.e * (math.log(1 / delta) + n_theta) / (epsilon * (math.e - 1)))
    return N


def lorenz_system(z, t, sigma=10.0, rho=28.0, beta=8 / 3):
    """Define the dynamics of the Lorenz system based on initial states

    :param z: The initial states of the Lorenz system (3 dimensions)
    :type z: numpy.ndarray
    :param t: The time at which the derivate of the system is computed
    :type t: int
    :param sigma: Lorenz system constant sigma, defaults to 10.
    :type sigma: float, optional
    :param rho: Lorenz system constant rho, defaults to 28.
    :type rho: float, optional
    :param beta: Lorenz system constant beta, defaults to 8/3
    :type beta: float, optional
    :return: The partial derivatives of the Lorenz system along x, y, and z
    :rtype: list
    """
    dzdt = [
        sigma * (z[1] - z[0]),
        z[0] * (rho - z[2]) - z[1],
        z[0] * z[1] - beta * z[2],
    ]
    return dzdt


def sample_lorenz(timesteps=100):
    """Obtain sample from Lorenz system over specified number of timesteps

    :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
    :type timesteps: int, optional
    :return: The sample from the Lorenz system at the last timestep
    :rtype: numpy.ndarray
    """
    t = np.linspace(0, timesteps, 1001)
    ru = default_rng().uniform
    z0 = np.array([ru(0, 1), ru(0, 1), ru(0, 1)])
    sol = odeint(lorenz_system, z0, t)
    return sol[-1]


def make_sample_n(sample_fn, parallel=True, pool=None):
    """Takes in a sample function and allows for parallelized sampling

    :param sample_fn: A function to compute samples
    :type sample_fn: function
    :param parallel: True if parallelization is to be used to compute samples, defaults to True
    :type parallel: bool, optional
    :param pool: Pool to use for parallelization if specified, defaults to None
    :type pool: multiprocessing.pool.Pool, optional
    """

    def sample_n(n, pool=pool):
        """Inner function to draw n samples using a specified sample function

        :param n: The number of samples to compute
        :type n: int
        :param pool: Pool to use for parallelization if specified, defaults to pool
        :type pool: multiprocessing.pool.Pool, optional
        :return: Array of n samples
        :rtype: numpy.ndarray
        """
        if parallel:
            if pool is None:
                print(f"Using {cpu_count()} CPUs")
                p = Pool(cpu_count())
            else:
                p = pool
            return np.array(list(p.map(sample_fn, np.arange(n))))
        else:
            return np.array([sample_fn() for i in range(n)])

    return sample_n


sample_lorenz_n = make_sample_n(sample_lorenz)
"""Function to sample from Lorenz system

:param n: The number of samples to compute.
:type n: int
"""



import cvxpy as cp
import numpy as np
import time

from functools import partial
from scipy.optimize import minimize_scalar


def format_time(t):
    """Formats time into string with minutes and seconds

    :param t: The time in seconds
    :type t: float
    :return: Formatted string with minutes and seconds
    :rtype: string
    """
    time_string = time.strftime("%M:%S", time.gmtime(t))
    time_list = time_string.split(":")
    num_min = time_list[0]
    num_sec = time_list[1]
    return f"{num_min} minutes and {num_sec} seconds"


def solve_p_norm(sample, n_x=3, p=2, const=None):
    """Solves the scenario relaxation problem for the given sample with p-Norm Balls

    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param n_x: The state dimension, defaults to 3
    :type n_x: int, optional
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :param const: The constraints placed on the parameters A and b, defaults to None
    :type const: string, optional
    :return: The values of matrix A and vector b corresponding to the optimal p-Norm Ball, as well as the status of the optimizer.
    :rtype: tuple
    """
    if const is None:
        A = cp.Variable((n_x, n_x), symmetric=True)
        b = cp.Variable((n_x, 1))
    elif const == "diagonal":
        a = cp.Variable((n_x, 1))
        A = cp.diag(a)
        b = cp.Variable((n_x, 1))
    elif const == "scalar":
        sigma = cp.Variable()
        A = sigma * np.identity(n_x)
        b = np.zeros((3, 1))

    obj = cp.Minimize(-cp.log_det(A))
    constraints = [cp.pnorm(A @ r.reshape(3, 1) - b, p=p) <= 1 for r in sample]
    prob = cp.Problem(obj, constraints)
    prob.solve()

    if const != "scalar":
        return A.value, b.value, prob.status
    else:
        return A, b, prob.status


def p_norm_cont(arr, axis, default_val, n_x, A_val, b_val, p, minimum=True):
    """Solve for the optimal value that satisfies the p-Norm Ball conditions at the specified axis

    :param arr: Array of shape (n_x - 1,) containing the independent variables of the p-norm condition
    :type arr: numpy.ndarray
    :param axis: The axis of the dependent variable for which to solve for (i.e. z -> axis=2).
    :type axis: int
    :param default_val: The value to return if no solution for the dependent variable is found that satisfies the p-norm conditions
    :type default_val: float
    :param n_x: The state dimension
    :type n_x: int
    :param A_val: The matrix of shape (n_x, n_x) corresponding to the optimal p-norm ball
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (n_x, 1) corresponding to the optimal p-norm ball
    :type b_val: numpy.ndarray
    :param p: The order of p-norm
    :type p: int
    :param minimum: True if optimizing for the minimal value of the dependent variable that satisfies the p-norm conditions, defaults to True
    :type minimum: bool, optional
    :return: The value at the specified axis which corresponds the the optimal value of the (n_x, 1) vector that satisfies the p-Norm Ball conditions at the specified axis
    :rtype: float
    """
    vec = cp.Variable((n_x, 1))
    other_dims = list(range(n_x))
    other_dims.remove(axis)
    constraints = [vec[i][0] == arr[j] for i, j in zip(other_dims, range(n_x - 1))]
    constraints.append(cp.pnorm(A_val @ vec - b_val, p=p) <= 1)
    if minimum:
        obj = cp.Minimize(vec[axis])
    else:
        obj = cp.Maximize(vec[axis])
    prob = cp.Problem(obj, constraints)

    try:
        prob.solve()
    except:
        return default_val

    if prob.status != "optimal":
        return default_val

    return vec.value[axis]


def p_norm_cont_proj(arr, axis, default_val, n_x, A_val, b_val, p):
    """Minimizes the p-Norm value with respect to value at the specified axis.

    :param arr: Array of shape (n_x - 1,) containing the independent variables of the p-norm condition.
    :type arr: numpy.ndarray
    :param axis: The axis of the dependent variable for which to solve for (i.e. z -> axis=2).
    :type axis: int
    :param default_val: The value to return if no solution for the dependent variable is found that satisfies the p-norm conditions.
    :type default_val: float
    :param n_x: The state dimension.
    :type n_x: int
    :param A_val: The matrix of shape (n_x, n_x) corresponding to the optimal p-norm ball.
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (n_x, 1) corresponding to the optimal p-norm ball.
    :type b_val: numpy.ndarray
    :param p: The order of p-norm.
    :type p: int
    :return: The value at the specified axis which corresponds the the minimum p-Norm value of the (n_x, 1)  vector.
    :rtype: float
    """
    vec = np.zeros((3))
    other_dims = list(range(n_x))
    other_dims.remove(axis)
    for i, j in zip(other_dims, range(n_x - 1)):
        vec[i] = arr[j]

    def f(x):
        vec[axis] = x
        return np.linalg.norm(A_val @ vec.reshape((3, 1)) - b_val, ord=p)

    res = minimize_scalar(f)
    vec[axis] = res.x

    if np.linalg.norm(A_val @ vec.reshape((3, 1)) - b_val, ord=p) <= 1:
        return res.x
    else:
        return default_val


def compute_contour_2D(sample, A_val, b_val, cont_axis=2, n_x=3, p=2, grid_n=200):
    """Computes the 3D contour for 2 dimensions based on sample data and the A_val, and b_val corresponding to the optimal p-norm ball.

    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param A_val: The matrix of shape (n_x, n_x) corresponding to the optimal p-norm ball
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (n_x, 1) corresponding to the optimal p-norm ball
    :type b_val: numpy.ndarray
    :param cont_axis: The axis for which the contours are to be solved for, defaults to 2
    :type cont_axis: int, optional
    :param n_x: The state dimension, defaults to 3
    :type n_x: int, optional
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 200
    :type grid_n: int, optional
    :return: The meshgrid, corresponding computed contour, and the extremum values for the chosen axis
    :rtype: tuple
    """
    x_min, x_max = sample[:, 0].min(), sample[:, 0].max()
    y_min, y_max = sample[:, 1].min(), sample[:, 1].max()
    z_min, z_max = sample[:, 2].min(), sample[:, 2].max()

    x = np.linspace(
        x_min - 0.4 * (x_max - x_min), x_max + 0.4 * (x_max - x_min), grid_n
    )
    y = np.linspace(
        y_min - 0.4 * (y_max - y_min), y_max + 0.4 * (y_max - y_min), grid_n
    )
    z = np.linspace(
        x_min - 0.4 * (z_max - z_min), z_max + 0.4 * (z_max - z_min), grid_n
    )

    if cont_axis == 2:
        d0, d1 = np.meshgrid(x, y)
        c_min, c_max = z_min, z_max
    elif cont_axis == 1:
        d0, d1 = np.meshgrid(x, z)
        c_min, c_max = y_min, y_max
    elif cont_axis == 0:
        d0, d1 = np.meshgrid(y, z)
        c_min, c_max = x_min, x_max

    d2 = np.array([d0.flatten(), d1.flatten()]).T

    solve_cont_d2 = partial(
        p_norm_cont_proj,
        axis=cont_axis,
        default_val=c_max + 1,
        n_x=n_x,
        A_val=A_val,
        b_val=b_val,
        p=p,
    )
    cont = np.fromiter(map(solve_cont_d2, d2), dtype=np.float64).reshape(grid_n, grid_n)

    return d0, d1, cont, c_min, c_max


def compute_contour_3D(sample, A_val, b_val, cont_axis=2, n_x=3, p=2, grid_n=200):
    """Computes the 3D contour for 3 dimensions based on sample data and the A_val, and b_val corresponding to the optimal p-norm ball.

    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param A_val: The matrix of shape (n_x, n_x) corresponding to the optimal p-norm ball
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (n_x, 1) corresponding to the optimal p-norm ball
    :type b_val: numpy.ndarray
    :param cont_axis: The axis for which the contours are to be solved for, defaults to 2
    :type cont_axis: int, optional
    :param n_x: The state dimension, defaults to 3
    :type n_x: int, optional
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 200
    :type grid_n: int, optional
    :param minimum: True if optimizing for the minimal value of the dependent variable that satisfies the p-norm conditions, defaults to True
    :type minimum: bool, optional
    :return: The meshgrid, corresponding computed contour, and the extremum values for the chosen axis
    :rtype: tuple
    """
    x_min, x_max = sample[:, 0].min(), sample[:, 0].max()
    y_min, y_max = sample[:, 1].min(), sample[:, 1].max()
    z_min, z_max = sample[:, 2].min(), sample[:, 2].max()

    x = np.linspace(
        x_min - 0.4 * (x_max - x_min), x_max + 0.4 * (x_max - x_min), grid_n
    )
    y = np.linspace(
        y_min - 0.4 * (y_max - y_min), y_max + 0.4 * (y_max - y_min), grid_n
    )
    z = np.linspace(
        x_min - 0.4 * (z_max - z_min), z_max + 0.4 * (z_max - z_min), grid_n
    )

    if cont_axis == 2:
        d0, d1 = np.meshgrid(x, y)
        c_min, c_max = z_min, z_max
    elif cont_axis == 1:
        d0, d1 = np.meshgrid(x, z)
        c_min, c_max = y_min, y_max
    elif cont_axis == 0:
        d0, d1 = np.meshgrid(y, z)
        c_min, c_max = x_min, x_max

    d2 = np.array([d0.flatten(), d1.flatten()]).T

    solve_cont_d2_min = partial(
        p_norm_cont,
        axis=cont_axis,
        default_val=c_max + 1,
        n_x=n_x,
        A_val=A_val,
        b_val=b_val,
        p=p,
        minimum=True,
    )

    solve_cont_d2_max = partial(
        p_norm_cont,
        axis=cont_axis,
        default_val=c_min - 1,
        n_x=n_x,
        A_val=A_val,
        b_val=b_val,
        p=p,
        minimum=False,
    )

    cont_min = np.fromiter(map(solve_cont_d2_min, d2), dtype=np.float64).reshape(
        grid_n, grid_n
    )
    cont_max = np.fromiter(map(solve_cont_d2_max, d2), dtype=np.float64).reshape(
        grid_n, grid_n
    )

    return d0, d1, cont_min, cont_max, c_min, c_max


def empirical_estimate(samples, A_val, b_val, n_x=3, p=2):
    """Computes the ratio of samples within the estimated reachable set

    :param samples: Sample from dynamical system (num_samples, n_x)
    :type samples: numpy.ndarray
    :param A_val: The matrix of shape (n_x, n_x) corresponding to the optimal p-norm ball
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (n_x, 1) corresponding to the optimal p-norm ball
    :type b_val: numpy.ndarray
    :param n_x: The state dimension, defaults to 3
    :type n_x: int, optional
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :return: The ratio of samples within the estimated reachability set
    :rtype: float
    """
    num_samples = samples.shape[0]
    count = 0
    for sample in samples:
        vec = sample.reshape(n_x, 1)
        if np.linalg.norm(A_val @ vec - b_val, ord=p) <= 1:
            count += 1
    return count / num_samples



from functools import partial
import numpy as np
import time
import warnings

from dyn_sys import DynamicSystem
from graph_utils import plot_contour_2D, plot_contour_3D, plot_sample
from sampling import num_samples
from utils import (
    compute_contour_2D,
    compute_contour_3D,
    empirical_estimate,
    format_time,
    solve_p_norm,
)


class Estimator:
    """Class which allows for estimation of the reachable sets for a given dynamical system.

    :param dyn_sys: An instance of :class:`dadra.DynamicSystem`
    :type dyn_sys: DynamicSystem
    :param epsilon: The accuracy parameter, defaults to 0.05
    :type epsilon: float, optional
    :param delta: The confidence parameter, defaults to 1e-9
    :type delta: float, optional
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :param const: The constraints placed on the parameters A and b, defaults to None
    :type const: string, optional
    :param normalize: If true, the sample is normalized, defaults to True
    :type normalize: bool, optional
    """

    def __init__(
        self,
        dyn_sys: DynamicSystem,
        epsilon=0.05,
        delta=1e-9,
        p=2,
        const=None,
        normalize=True,
    ):

        self.dyn_sys = dyn_sys
        self.check_sys()

        self.epsilon = epsilon
        self.delta = delta
        self.p = p
        self.const = const
        self.normalize = normalize
        self.num_samples = self.get_num_samples()
        self.samples = self.get_sample()

        self.A, self.b, self.status = self.solve_p()

    @classmethod
    def estimator_from_func(
        cls,
        dyn_func,
        intervals,
        state_dim,
        timesteps=100,
        parts=1001,
        epsilon=0.05,
        delta=1e-9,
        p=2,
        const=None,
        normalize=True,
    ):
        """Class method that allows for an instance of :class:`dadra.Estimator` to be initialized
        using a dynamic function, and the components for an instance of :class:`DynamicSystem`
        rather than explicitly passing in a :class:`DynamicSystem` object

        :param dyn_func: The function defining the dynamics of the system
        :type dyn_func: function
        :param intervals: The intervals corresponding to the possible values of the initial states of the variables in the system
        :type intervals: list
        :param state_dim: The degrees of freedom of the system
        :type state_dim: int
        :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
        :type timesteps: int, optional
        :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 1001
        :type parts: int, optional
        :param epsilon: The accuracy parameter, defaults to 0.05
        :type epsilon: float, optional
        :param delta: The confidence parameter, defaults to 1e-9
        :type delta: float, optional
        :param p: The order of p-norm, defaults to 2
        :type p: int, optional
        :param const: The constraints placed on the parameters A and b, defaults to None
        :type const: string, optional
        :param normalize: If true, the sample is normalized, defaults to True
        :type normalize: bool, optional
        :return: A :class:`dadra.Estimator` object
        :rtype: :class:`dadra.Estimator`
        """
        dyn_sys = DynamicSystem(dyn_func, intervals, state_dim, timesteps, parts)
        return cls(dyn_sys, epsilon, delta, p, const, normalize)

    @classmethod
    def estimator_from_list(
        cls,
        func_list,
        intervals,
        timesteps=100,
        parts=1001,
        epsilon=0.05,
        delta=1e-9,
        p=2,
        const=None,
        normalize=True,
    ):
        """Class method that allows for an instance of :class:`dadra.Estimator` to be initialized
        using a list of dynamic functions, and the components for an instance of :class:`DynamicSystem`
        rather than explicitly passing in a :class:`DynamicSystem` object

        :param dyn_func: The list of functions, one for each variable, that define the dynamics of the system
        :type dyn_func: function
        :param intervals: The intervals corresponding to the possible values of the initial states of the variables in the system
        :type intervals: list
        :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
        :type timesteps: int, optional
        :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 1001
        :type parts: int, optional
        :param epsilon: The accuracy parameter, defaults to 0.05
        :type epsilon: float, optional
        :param delta: The confidence parameter, defaults to 1e-9
        :type delta: float, optional
        :param p: The order of p-norm, defaults to 2
        :type p: int, optional
        :param const: The constraints placed on the parameters A and b, defaults to None
        :type const: string, optional
        :param normalize: If true, the sample is normalized, defaults to True
        :type normalize: bool, optional
        :return: A :class:`dadra.Estimator` object
        :rtype: :class:`dadra.Estimator`
        """
        dyn_sys = DynamicSystem.get_system(func_list, intervals, timesteps, parts)
        return cls(dyn_sys, epsilon, delta, p, const, normalize)

    def check_sys(self):
        """Checks whether the dynamic system passed to the constructor is valid

        :raises TypeError: If an object of type other than :class:`dadra.DynamicSystem` is passed
        """
        if not isinstance(self.dyn_sys, DynamicSystem):
            raise TypeError("Object of type DynamicSystem must be passed")

    def get_num_samples(self):
        """Compute the number of samples needed to satisfy the specified probabilistic guarantees

        :return: The number of samples needed to satisfy the specified probabilistic guarantees
        :rtype: int
        """
        n_x = self.dyn_sys.state_dim
        return num_samples(self.epsilon, self.delta, n_x, self.const)

    def get_sample(self):
        """Draws the number of samples necessary to satisfy the specified probabilistic guarantees

        :return: Array of samples
        :rtype: numpy.ndarray
        """
        print(f"Drawing {self.num_samples} samples")
        start_sampling = time.perf_counter()
        samples = self.dyn_sys.sample_system(self.num_samples)
        end_sampling = time.perf_counter()

        print(
            f"Time to draw {self.num_samples} samples: {format_time(end_sampling - start_sampling)}"
        )

        if self.normalize:
            samples = (samples - np.mean(samples, axis=0)) / np.std(samples)
        return samples

    def solve_p(self):
        """Solves for the optimal p-norm ball that estimates the reachable set

        :return: The matrix corresponding to parameter A of the p-norm ball, the vector corresponding to parameter b of the p-norm ball, and the status of the optimization problem
        :rtype: tuple
        """
        print(f"Solving for optimal p-norm ball (p={self.p})")
        start_time = time.perf_counter()
        A, b, status = solve_p_norm(
            self.samples, self.dyn_sys.state_dim, self.p, self.const
        )
        end_time = time.perf_counter()
        if status != "optimal":
            warnings.warn("Failed to find optimal solution for p-norm ball")
        else:
            print(
                f"Time to solve for optimal p-norm ball: {format_time(end_time - start_time)}"
            )
        return A, b, status

    def empirical_estimate(self, num_samples_emp=None):
        """Computes the ratio of samples within the estimated reachable set

        :param num_samples_emp: The number of samples to draw for the empirical estimate, defaults to the number of samples drawn to obtain the p-norm ball
        :type num_samples_emp: int, optional
        """
        if self.status != "optimal":
            warnings.warn(
                "Non-optimal solution to p-norm ball. Empirical estimate may be innaccurate"
            )

        if num_samples_emp is None:
            num_samples_emp = self.num_samples

        print(f"Drawing {num_samples_emp} samples for empirical estimate")
        start_sampling = time.perf_counter()
        samples_emp = self.dyn_sys.sample_system(num_samples_emp)
        end_sampling = time.perf_counter()

        print(
            f"Time to draw {num_samples_emp} samples: {format_time(end_sampling - start_sampling)}"
        )

        if self.normalize:
            samples_emp = (samples_emp - np.mean(samples_emp, axis=0)) / np.std(
                samples_emp
            )

        ratio = empirical_estimate(
            samples_emp, self.A, self.b, self.dyn_sys.state_dim, self.p
        )
        print(f"Ratio of samples within the estimated reachability set: {ratio}")

    def summary(self):
        """Prints a summary of the attributes of this instance of :class:`Estimator`"""
        summary_str = (
            "----------------------------------------------------------------" + "\n"
        )
        summary_str += "Estimator Summary" + "\n"
        summary_str += (
            "================================================================" + "\n"
        )
        summary_str += f"State dimension: {self.dyn_sys.state_dim}" + "\n"
        summary_str += f"Accuracy parameter epsilon: {self.epsilon}" + "\n"
        summary_str += f"Confidence parameter delta: {self.delta}" + "\n"
        summary_str += f"p-norm p value: {self.p}" + "\n"
        summary_str += f"Constraints on p-norm ball: {self.const}" + "\n"
        summary_str += f"Number of samples: {self.num_samples}" + "\n"
        summary_str += f"Status of p-norm ball solution: {self.status}" + "\n"
        summary_str += (
            "----------------------------------------------------------------" + "\n"
        )
        print(summary_str)

    def plot_samples(self, fig_name):
        """Plots the samples of of shape (num_items, 3) in 3D

        :param fig_name: The name of the file to save the plot to
        :type fig_name: string
        """
        plot_sample(self.samples, fig_name)

    def plot_2D_cont(self, fig_name, grid_n=200):
        """Computes the contours of the reachable set and plots them in 2D

        :param fig_name: The name of the file to save the plot to
        :type fig_name: string
        :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 200
        :type grid_n: int
        """
        cont_compute = partial(
            compute_contour_2D,
            sample=self.samples,
            A_val=self.A,
            b_val=self.b,
            n_x=self.dyn_sys.state_dim,
            p=self.p,
            grid_n=grid_n,
        )

        print("Computing 2D contours")
        cont_start = time.perf_counter()
        xv1, yv1, z_cont, z_min, z_max = cont_compute(cont_axis=2)
        xv2, zv1, y_cont, y_min, y_max = cont_compute(cont_axis=1)
        yv2, zv2, x_cont, x_min, x_max = cont_compute(cont_axis=0)
        cont_end = time.perf_counter()
        print(f"Time to compute 2D contours: {format_time(cont_end - cont_start)}")

        plot_contour_2D(
            xv1,
            yv1,
            z_cont,
            z_max,
            self.samples,
            fig_name,
            xv2=xv2,
            zv1=zv1,
            y_cont=y_cont,
            y_level=y_max,
            yv2=yv2,
            zv2=zv2,
            x_cont=x_cont,
            x_level=x_max,
        )

    def plot_3D_cont(self, fig_name, grid_n=100, gif_name=None):
        """Computes and plots the contours in 3D with the option for saving an animated gif of the rotating graph

        :param fig_name: The name of the file to save the plot to
        :type fig_name: string
        :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 100
        :type grid_n: int
        :param gif_name: The name of the file to save the gif to, defaults to None
        :type gif_name: string, optional
        """
        print("Computing 3D contour")
        cont_start = time.perf_counter()
        d0, d1, cont_min, cont_max, c_min, c_max = compute_contour_3D(
            self.samples, self.A, self.b, grid_n=grid_n
        )
        cont_end = time.perf_counter()
        print(f"Time to compute 3D contour: {format_time(cont_end - cont_start)}")

        plot_contour_3D(
            d0,
            d1,
            cont_min,
            c_min,
            c_max,
            self.samples,
            fig_name,
            gif_name=gif_name,
            z_cont2=cont_max,
        )



from setuptools import find_packages, setup
import setuptools

setup(
    name="dadra",
    packages=setuptools.find_packages(),
    license="MIT",
    version="0.0.2",
    description="Library for Data-Driven Reachability Analysis",
    author="Jared Mejia",
    author_email="jaredmejia@berkeley.edu",
    url="https://github.com/JaredMejia/dadra",
    install_requires=["cvxpy", "matplotlib", "numpy", "scipy"],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ),
)



import numpy as np

from collections.abc import Callable
from inspect import signature, Parameter
from numpy.random import default_rng
from sampling import make_sample_n
from scipy.integrate import odeint


class DynamicSystem:
    """Class implementation of a dynamical system that allows for parallelized sampling.

    :param dyn_func: The function defining the dynamics of the system
    :type dyn_func: function
    :param intervals: The intervals corresponding to the possible values of the initial states of the variables in the system
    :type intervals: list
    :param state_dim: The degrees of freedom of the system
    :type state_dim: int
    :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
    :type timesteps: int, optional
    :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 1001
    :type parts: int, optional
    """

    def __init__(self, dyn_func, intervals, state_dim, timesteps=100, parts=1001):
        """Constructor method"""
        self.dyn_func = dyn_func
        self.intervals = intervals
        self.state_dim = state_dim
        self.timesteps = timesteps
        self.parts = parts

        self.check_intervals()
        self.check_dyn_func()

    @classmethod
    def get_system(cls, func_list, intervals, timesteps=100, parts=1001):
        """Class method that allows for an instance of :class:`dadra.DynamicSystem` to be initialized
        using a list of functions, one for each variable, to define the dynamics of a system rather than
        a single function

        :param func_list: The list of functions, one for each variable, that define the dynamics of the system
        :type func_list: list
        :param intervals: The intervals corresponding to the possible values of the initial states of the variables in the system
        :type intervals: list
        :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
        :type timesteps: int, optional
        :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 1001
        :type parts: int, optional
        :raises TypeError: If not all functions in ``func_list`` are callable
        :raises ValueError: If not all functions in ``func_list`` include parameters for the intial state and time
        :raises ValueError: If not all functions in ``func_list`` include parameters for the intial state, time, and the provided additional arguments
        :return: A :class:`dadra.DynamicSystem` object
        :rtype: :class:`dadra.DynamicSystem`
        """
        state_dim = len(func_list)
        num_intervs = len(intervals)
        num_extra_intervs = num_intervs - state_dim

        for func in func_list:
            if not isinstance(func, Callable):
                raise TypeError(
                    "There must be a callable function defined for each variable"
                )
            sig = signature(func)
            num_params = len(sig.parameters)

            if num_extra_intervs == 0 and num_params != 2:
                raise ValueError(
                    "Each dynamic function must be a function of the initial state and time"
                )
            elif num_extra_intervs > 0 and num_params != (2 + num_extra_intervs):
                raise ValueError(
                    "Each dynamic function must be a function of the initial state, time, and the provided additional arguments"
                )

        global dyn_to_sys  # note: can only run experiments sequentially, create new object each time

        def dyn_to_sys(arr, t, *args):
            if len(args) != 0:
                derivs = [f(arr, t, *args) for f in func_list]
            else:
                derivs = [f(arr, t) for f in func_list]
            return derivs

        dyn_func = dyn_to_sys

        dyn_sys = cls(dyn_func, intervals, state_dim, timesteps, parts)
        return dyn_sys

    def check_intervals(self):
        """Checks whether the list of intervals in this instance of the :class:`dadra.DynamicSystem` class
        is valid

        :raises ValueError: If there is not an interval defined for each variable
        :raises ValueError: If not all intervals include an upper and a lower bound
        :raises ValueError: If not all intervals consist of two numbers
        """
        if len(self.intervals) < self.state_dim:
            raise ValueError("There must be an interval defined for each variable")

        for interval in self.intervals:
            if len(interval) != 2:
                raise ValueError(
                    "Each interval must include an upper and a lower bound"
                )
            for v in interval:
                if type(v) == int or type(v) == float:
                    pass
                else:
                    raise ValueError("Each interval must consist of two numbers")

    def check_dyn_func(self):
        """Checks whether the dynamic function in this instance of the :class:`dadra.DynamicSystem` class
        is valid

        :raises TypeError: If the provided dynamic function is not callable
        :raises ValueError: If there is an invalid number of required arguments in the dynamic function
        """
        if not isinstance(self.dyn_func, Callable):
            raise TypeError("A callable dynamic function must be provided")

        sig = signature(self.dyn_func)
        num_params = len(sig.parameters)
        num_required = num_params

        for p in sig.parameters.values():
            if p.default != Parameter.empty or p.kind == Parameter.VAR_POSITIONAL:
                num_required -= 1

        num_extra_intervs = len(self.intervals) - self.state_dim
        if num_required > (2 + num_extra_intervs):
            raise ValueError("Invalid number of required arguments in dynamic function")

    def system_sampler(self, x=None):
        """Obtains sample from specified system

        :param x: Placeholder variable over which to evaluate, defaults to None
        :type x: NoneType, optional
        :return: The sample from the specified system at the last timestep
        :rtype: numpy.ndarray
        """
        t = np.linspace(0, self.timesteps, self.parts)
        ru = default_rng().uniform
        rand_intervals = [ru(lower, upper) for lower, upper in self.intervals]
        initial_state = np.array(rand_intervals[: self.state_dim])
        extra_intervs = tuple(rand_intervals[self.state_dim :])
        if len(extra_intervs) != 0:
            sol = odeint(self.dyn_func, initial_state, t, args=extra_intervs)
        else:
            sol = odeint(self.dyn_func, initial_state, t)
        return sol[-1]

    def sample_system(self, N):
        """Draws ``N`` samples from the specified system

        :param N: The number of samples to be drawn
        :type N: int
        :return: Array of ``N`` samples
        :rtype: numpy.ndarray
        """
        return make_sample_n(self.system_sampler)(N)



import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np


def plot_sample(sample, fig_name):
    """Plots a sample of shape (num_items, 3) in 3D

    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param fig_name: The name of the file to save the plot to
    :type fig_name: string
    """
    fig, axs = plt.subplots(3, figsize=(5, 15))
    axs[0].plot(sample[:, 0], sample[:, 1], "k.")
    axs[0].set_xlabel("x")
    axs[0].set_ylabel("y")
    axs[1].plot(sample[:, 0], sample[:, 2], "k.")
    axs[1].set_xlabel("x")
    axs[1].set_ylabel("z")
    axs[2].plot(sample[:, 1], sample[:, 2], "k.")
    axs[2].set_xlabel("y")
    axs[2].set_ylabel("z")
    plt.savefig(fig_name)


def plot_contour_2D(
    xv1,
    yv1,
    z_cont,
    z_level,
    sample,
    fig_name,
    xv2=None,
    zv1=None,
    y_cont=None,
    y_level=None,
    yv2=None,
    zv2=None,
    x_cont=None,
    x_level=None,
):
    """Plots the contours provided (output of utils.compute_contour) in 2D

    :param xv1: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension
    :type xv1: numpy.ndarray
    :param yv1: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension
    :type yv1: numpy.ndarray
    :param z_cont: A (grid_n, grid_n) matrix, each point a function of xv1 and yv1
    :type z_cont: numpy.ndarray
    :param z_level: The z-value at which the level set is to be drawn
    :type z_level: float
    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param fig_name: The name of the file to save the plot to
    :type fig_name: string
    :param xv2: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension, defaults to None
    :type xv2: numpy.ndarray, optional
    :param zv1: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension, defaults to None
    :type zv1: numpy.ndarray, optional
    :param y_cont: A (grid_n, grid_n) matrix, each point a function of xv2 and zv1, defaults to None
    :type y_cont: numpy.ndarray, optional
    :param y_level: The y-value at which the level set is to be drawn, defaults to None
    :type y_level: float, optional
    :param yv2: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension, defaults to None
    :type yv2: numpy.ndarray, optional
    :param zv2: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension, defaults to None
    :type zv2: numpy.ndarray, optional
    :param x_cont: A (grid_n, grid_n) matrix, each point a function of yv2 and zv2, defaults to None
    :type x_cont: numpy.ndarray, optional
    :param x_level: The x-value at which the level set is to be drawn, defaults to None
    :type x_level: float, optional
    """
    if y_cont is not None and x_cont is not None:
        fig, axs = plt.subplots(3, figsize=(5, 15))

    elif y_cont is not None:
        fig, axs = plt.subplots(2, figsize=(5, 10))

    else:
        fig, axs = plt.subplots(1, figsize=(5, 5))
        axs.plot(sample[:, 0], sample[:, 1], "k.")
        axs.contour(xv1, yv1, z_cont, levels=[z_level], colors="b")
        axs.set_xlabel("x")
        axs.set_ylabel("y")
        axs.set_title("contour in x-y plane")
        plt.show()
        return

    axs[0].plot(sample[:, 0], sample[:, 1], "k.")
    axs[0].contour(xv1, yv1, z_cont, levels=[z_level], colors="b")
    axs[0].set_xlabel("x")
    axs[0].set_ylabel("y")
    axs[0].set_title("contour in x-y plane")

    if y_cont is not None:
        axs[1].plot(sample[:, 0], sample[:, 2], "k.")
        axs[1].contour(xv2, zv1, y_cont, levels=[y_level], colors="b")
        axs[1].set_xlabel("x")
        axs[1].set_ylabel("z")
        axs[1].set_title("contour in x-z plane")

        if x_cont is not None:
            axs[2].plot(sample[:, 1], sample[:, 2], "k.")
            axs[2].contour(yv2, zv2, x_cont, levels=[x_level], colors="b")
            axs[2].set_xlabel("y")
            axs[2].set_ylabel("z")
            axs[2].set_title("contour in y-z plane")

    plt.savefig(fig_name)


def plot_contour_3D(
    xv,
    yv,
    z_cont,
    z_min,
    z_max,
    sample,
    fig_name,
    gif_name=None,
    z_cont2=None,
):
    """Plots the contours in 3D with the option for saving an animated gif of the rotating graph

    :param xv: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension
    :type xv: numpy.ndarray
    :param yv: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension
    :type yv: numpy.ndarray
    :param z_cont: A (grid_n, grid_n) matrix, each point a function of xv1 and yv1 (corresponding to the maximum optimal solutions that satisfy the p-norm)
    :type z_cont: numpy.ndarray
    :param z_min: The minimum z-value
    :type z_min: float
    :param z_max: The maximum z-value
    :type z_max: float
    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param fig_name: The name of the file to save the plot to
    :type fig_name: string
    :param gif_name: The name of the file to save the gif to, defaults to None
    :type gif_name: string, optional
    :param z_cont2: A (grid_n, grid_n) matrix, each point a function of xv and yv (corresponding to the minimum optimal solutions that satisfy the p-norm), defaults to None
    :type z_cont2: numpy.ndarray, optional
    """
    fig = plt.figure(figsize=(10, 10))
    ax = plt.axes(projection="3d")
    v_min = z_min - 0.25 * (z_max - z_min)
    v_max = z_max + 0.25 * (z_max - z_min)
    ax.scatter3D(
        sample[:, 0],
        sample[:, 1],
        sample[:, 2],
        c=sample[:, 2],
        cmap="binary",
        vmin=v_min,
        vmax=z_max,
    )
    if z_cont2 is None:
        ax.contour3D(
            xv, yv, z_cont, 50, cmap="Blues", vmin=v_min, vmax=v_max, alpha=0.8
        )
    else:
        abs_min = np.min(z_cont)
        abs_max = np.max(z_cont2)
        abs_avg = (abs_min + abs_max) / 2
        levels_1 = list(np.linspace(abs_min, abs_avg, 25))
        levels_2 = list(np.linspace(abs_avg, abs_max, 25))
        ax.contour3D(
            xv,
            yv,
            z_cont,
            50,
            cmap="Blues",
            vmin=v_min,
            vmax=v_max,
            levels=levels_1,
        )
        ax.contour3D(
            xv,
            yv,
            z_cont2,
            50,
            cmap="Blues",
            vmin=v_min,
            vmax=v_max,
            levels=levels_2,
        )

    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")
    ax.set_title("3D contour")

    if gif_name is not None:

        def rotate(angle):
            ax.view_init(azim=angle)

        rot_animation = animation.FuncAnimation(
            fig, rotate, frames=np.arange(0, 362, 2), interval=100
        )
        rot_animation.save(gif_name, dpi=100)

    plt.savefig(fig_name)
